@@ -54,6 +54,9 @@ class TestResponseCallbacks(object):
         r['a'] = 'foo'
         assert r['a'] == 'static'
 
+    def test_case_insensitive_command_names(self, r):
+        assert r.response_callbacks['del'] == r.response_callbacks['DEL']
+
 
 class TestRedisCommands(object):
 
@@ -99,20 +102,21 @@ class TestRedisCommands(object):
     @skip_if_server_version_lt('2.6.9')
     def test_client_kill(self, r, r2):
         r.client_setname('redis-py-c1')
-        r2[0].client_setname('redis-py-c2')
-        r2[1].client_setname('redis-py-c3')
-        test_clients = [client for client in r.client_list()
-                        if client.get('name')
-                        in ['redis-py-c1', 'redis-py-c2', 'redis-py-c3']]
-        assert len(test_clients) == 3
+        r2.client_setname('redis-py-c2')
+        clients = [client for client in r.client_list()
+                   if client.get('name') in ['redis-py-c1', 'redis-py-c2']]
+        assert len(clients) == 2
 
-        resp = r.client_kill(test_clients[1].get('addr'))
-        assert isinstance(resp, bool) and resp is True
+        clients_by_name = dict([(client.get('name'), client)
+                                for client in clients])
 
-        test_clients = [client for client in r.client_list()
-                        if client.get('name')
-                        in ['redis-py-c1', 'redis-py-c2', 'redis-py-c3']]
-        assert len(test_clients) == 2
+        client_addr = clients_by_name['redis-py-c2'].get('addr')
+        assert r.client_kill(client_addr) is True
+
+        clients = [client for client in r.client_list()
+                   if client.get('name') in ['redis-py-c1', 'redis-py-c2']]
+        assert len(clients) == 1
+        assert clients[0].get('name') == 'redis-py-c1'
 
     @skip_if_server_version_lt('2.8.12')
     def test_client_kill_filter_invalid_params(self, r):
@@ -129,25 +133,44 @@ class TestRedisCommands(object):
             r.client_kill_filter(_type="caster")
 
     @skip_if_server_version_lt('2.8.12')
-    def test_client_kill_filter(self, r, r2):
+    def test_client_kill_filter_by_id(self, r, r2):
         r.client_setname('redis-py-c1')
-        r2[0].client_setname('redis-py-c2')
-        r2[1].client_setname('redis-py-c3')
-        test_clients = [client for client in r.client_list()
-                        if client.get('name')
-                        in ['redis-py-c1', 'redis-py-c2', 'redis-py-c3']]
-        assert len(test_clients) == 3
+        r2.client_setname('redis-py-c2')
+        clients = [client for client in r.client_list()
+                   if client.get('name') in ['redis-py-c1', 'redis-py-c2']]
+        assert len(clients) == 2
 
-        resp = r.client_kill_filter(_id=test_clients[1].get('id'))
-        assert isinstance(resp, int) and resp == 1
+        clients_by_name = dict([(client.get('name'), client)
+                                for client in clients])
 
-        resp = r.client_kill_filter(addr=test_clients[2].get('addr'))
-        assert isinstance(resp, int) and resp == 1
+        client_2_id = clients_by_name['redis-py-c2'].get('id')
+        resp = r.client_kill_filter(_id=client_2_id)
+        assert resp == 1
 
-        test_clients = [client for client in r.client_list()
-                        if client.get('name')
-                        in ['redis-py-c1', 'redis-py-c2', 'redis-py-c3']]
-        assert len(test_clients) == 1
+        clients = [client for client in r.client_list()
+                   if client.get('name') in ['redis-py-c1', 'redis-py-c2']]
+        assert len(clients) == 1
+        assert clients[0].get('name') == 'redis-py-c1'
+
+    @skip_if_server_version_lt('2.8.12')
+    def test_client_kill_filter_by_addr(self, r, r2):
+        r.client_setname('redis-py-c1')
+        r2.client_setname('redis-py-c2')
+        clients = [client for client in r.client_list()
+                   if client.get('name') in ['redis-py-c1', 'redis-py-c2']]
+        assert len(clients) == 2
+
+        clients_by_name = dict([(client.get('name'), client)
+                                for client in clients])
+
+        client_2_addr = clients_by_name['redis-py-c2'].get('addr')
+        resp = r.client_kill_filter(addr=client_2_addr)
+        assert resp == 1
+
+        clients = [client for client in r.client_list()
+                   if client.get('name') in ['redis-py-c1', 'redis-py-c2']]
+        assert len(clients) == 1
+        assert clients[0].get('name') == 'redis-py-c1'
 
     @skip_if_server_version_lt('2.6.9')
     def test_client_list_after_client_setname(self, r):
@@ -1640,6 +1663,19 @@ class TestRedisCommands(object):
         assert isinstance(mock_cluster_resp_slaves.cluster(
             'slaves', 'nodeid'), dict)
 
+    @skip_if_server_version_lt('3.0.0')
+    def test_readwrite(self, r):
+        assert r.readwrite()
+
+    @skip_if_server_version_lt('3.0.0')
+    def test_readonly_invalid_cluster_state(self, r):
+        with pytest.raises(exceptions.RedisError):
+            r.readonly()
+
+    @skip_if_server_version_lt('3.0.0')
+    def test_readonly(self, mock_cluster_resp_ok):
+        assert mock_cluster_resp_ok.readonly() is True
+
     # GEO COMMANDS
     @skip_if_server_version_lt('3.2.0')
     def test_geoadd(self, r):
@@ -1892,6 +1928,31 @@ class TestRedisCommands(object):
                         min_idle_time=0, message_ids=(message_id,),
                         justid=True) == [message_id]
 
+    @skip_if_server_version_lt('5.0.0')
+    def test_xclaim_trimmed(self, r):
+        # xclaim should not raise an exception if the item is not there
+        stream = 'stream'
+        group = 'group'
+
+        r.xgroup_create(stream, group, id="$", mkstream=True)
+
+        # add a couple of new items
+        sid1 = r.xadd(stream, {"item": 0})
+        sid2 = r.xadd(stream, {"item": 0})
+
+        # read them from consumer1
+        r.xreadgroup(group, 'consumer1', {stream: ">"})
+
+        # add a 3rd and trim the stream down to 2 items
+        r.xadd(stream, {"item": 3}, maxlen=2, approximate=False)
+
+        # xclaim them from consumer2
+        # the item that is still in the stream should be returned
+        item = r.xclaim(stream, group, 'consumer2', 0, [sid1, sid2])
+        assert len(item) == 2
+        assert item[0] == (None, None)
+        assert item[1][0] == sid2
+
     @skip_if_server_version_lt('5.0.0')
     def test_xdel(self, r):
         stream = 'stream'
